#!/usr/bin/env ruby
require 'tk'
require 'tkextlib/bwidget'

#############################
# setup root
root = TkRoot.new(:title=>'Ruby/Tk test window')

#############################
# each page of notebook
class ScriptPage < TkText
  PAGE_ID = ['page'.freeze, '00000']

  def _new_page_id
    id = PAGE_ID.join
    PAGE_ID[1].succ!
    id
  end
  private :_new_page_id

  def _set_tab_text
    if @fname.value == '-'
      @parent.itemconfigure(@page_id, :text=>'*scratch*')
    else
      @parent.itemconfigure(@page_id, :text=>File.basename(@fname.value))
    end
  end
  private :_set_tab_text

  def initialize(notebook, fileopen=true)
    @parent = notebook
    @page_id = _new_page_id.freeze
    @page = @parent.insert(:end, @page_id, :text=>'-', 
                           :raisecmd=>proc{@parent.see(@page_id)})

    @frame = @parent.get_frame(@page_id)
    @frame.borderwidth = 2
    @frame.relief = :ridge

    # create text frame
    f = TkFrame.new(@frame, :borderwidth=>2, 
                    :relief=>:ridge).pack(:fill=>:both, :expand=>true)
    @txt = TkText.new(f, :undo=>true, :maxundo=>-1, :autoseparators=>true){
      yscrollbar(TkScrollbar.new(f).pack(:side=>:right, :fill=>:y))
      pack(:side=>:left, :fill=>:both, :expand=>true)
    }
    @txt.focus

    @path = @txt.path

    # create filename frame
    @fname = TkVariable.new('-')
    f = TkFrame.new(@frame).pack(:fill=>:x)
    @label = TkLabel.new(f, :textvariable=>@fname, :anchor=>'w', 
                         :font=>{:size=>-12, :weight=>:bold}) {
      pack(:side=>:left, :fill=>:x, :padx=>5)
    }
    @btn = TkButton.new(f, :text=>'exec', :command=>proc{eval_file}){
      pack(:side=>:right, :padx=>3, :pady=>2)
    }

    Tk.update
    @parent.compute_size
    @parent.raise(@page_id)

     if fileopen
       open_file()
     else
       _set_tab_text
     end
  end

  # -------------------------------
  def id
    @page_id
  end

  # -------------------------------
  def open_file
    if @txt.modified?
      ret = Tk.messageBox(:icon=>'warning', :type=>'okcancel', 
                          :default=>'cancel',
                          :message=>"Buffer '#{@fname.value}' is modified. "+
                                    'Realy erase? ')
      return if ret == 'cancel'
    end

    fpath = Tk.getOpenFile(:filetypes=>[ ['Ruby Scripts', ['.rb', '.rbw']], 
                                         ['ALL Files', '*'] ])
    return if fpath.empty?

    begin
      src = IO.read(fpath)
    rescue => e
      Tk.messageBox(:icon=>'error', :type=>'ok', 
                    :message=>"Fail to read '#{fpath}'.\n#{e.message}")
    end
  
    @txt.value = src
    @txt.modified(false)
    @txt.edit_reset
    @txt.focus

    @txt.mark_set('insert', '1.0')

    @fname.value = fpath
    _set_tab_text
  end

  # --------------------------------
  def save_file
    return unless @txt.modified?

    if @fname.value == '-'
      ret = Tk.messageBox(:icon=>'warning', :type=>'okcancel', :default=>'ok',
                          :message=>'Filename is not assigned. Save it? ')
      save_as_file(@txt, @fname) if ret == 'ok'
      return
    end

    begin
      fid = open(@fname.value, 'w')
    rescue => e
      Tk.messageBox(:icon=>'error', :type=>'ok', 
                    :message=>"Fail to open '#{@fname.value}'.\n#{e.message}")
    end

    begin
      fid.print(@txt.value)
    rescue =>e
      Tk.messageBox(:icon=>'error', :type=>'ok', 
                    :message=>"Fail to write '#{@fname.value}'.\n#{e.message}")
    ensure
      fid.close
    end

    @txt.modified(false)
  end

  # --------------------------------
  def save_as_file
    fpath = Tk.getSaveFile(:filetypes=>[ ['Ruby Scripts', ['.rb', '.rbw']], 
                                         ['ALL Files', '*'] ])
    return if fpath.empty?

    begin
      fid = open(fpath, 'w')
    rescue => e
      Tk.messageBox(:icon=>'error', :type=>'ok', 
                    :message=>"Fail to open '#{fpath}'.\n#{e.message}")
    end

    begin
      fid.print(@txt.value)
    rescue =>e
      Tk.messageBox(:icon=>'error', :type=>'ok', 
                    :message=>"Fail to write '#{fpath}'.\n#{e.message}")
    ensure
      fid.close
    end

    @txt.modified(false)
    @fname.value = fpath
    _set_tab_text
  end

  # --------------------------------
  def eval_file
    system('ruby', '-e', @txt.value)
  end

  # --------------------------------
  def close
    if @txt.modified?
      ret = Tk.messageBox(:icon=>'warning', :type=>'okcancel', 
                          :default=>'cancel', 
                          :message=>"Buffer '#{@fname.value}' is modified. "+
                                    'Realy close? ')
      return if ret != 'ok'
    end

    @parent.delete(@page_id)
  end
end

####################################
# notebook
class ScriptNote < Tk::BWidget::NoteBook
  def initialize(*args)
    super(*args)
    @pages = {}
    page = ScriptPage.new(self, false)
    @pages[page.id] = page
  end

  def open_file
    page = ScriptPage.new(self)
    @pages[page.id] = page
  end

  def save_file
    id = tk_send('raise')
    page = @pages[id]
    page.save_file if page
  end

  def save_as_file
    id = tk_send('raise')
    page = @pages[id]
    page.eval_as_file if page
  end

  def eval_file
    id = tk_send('raise')
    page = @pages[id]
    page.eval_file if page
  end

  def close
    id = tk_send('raise')
    page = @pages[id]
    page.close if page
    if get_page(index(id)).empty?
      @pages.delete(id)
      tk_send('raise', self.pages[-1])
    end
  end

  def empty?
    @pages.empty?
  end

  def close_all
    self.pages.each{|id|
      page = @pages[id]
      page.close if page
      @pages.delete(id) if get_page(index(id)).empty?
    }
    @pages.empty?
  end
end

####################################
# main part
note = ScriptNote.new(root).pack

def quit(note)
  exit if note.close_all
end

# create menu
mspec = [
  [ ['File', 0], 
    ['Open', proc{note.open_file}, 0], 
    ['Save', proc{note.save_file}, 0], 
    ['Save As', proc{note.save_as_file}, 5], 
    ['Close', proc{note.close; exit if note.empty?}, 0], 
    '---', 
    ['Exec', proc{note.eval_file}, 0], 
    '---', 
    ['Quit', proc{quit(note)}, 0]
  ]
]
root.add_menubar(mspec)

# manage wm_protocol
root.protocol(:WM_DELETE_WINDOW){quit(note)}

# --------------------------------
# start eventloop
Tk.mainloop
